@@ -17,8 +17,9 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-import attr
 import enum
+
+import attr
 import tensorflow as tf
 
 
@@ -63,6 +64,8 @@ class AdvNeighborConfig(object):
       (FGSM) attack.
     pgd_epsilon: radius of the epsilon ball to project back to. Only used in
       Projected Gradient Descent (PGD) attack.
+    random_init: Apply a random perturbation before FGSM/PGD steps. Default set
+      to `False` for no random initialization being applied.
   """
   feature_mask = attr.ib(default=None)
   adv_step_size = attr.ib(default=0.001)
@@ -71,6 +74,7 @@ class AdvNeighborConfig(object):
   clip_value_max = attr.ib(default=None)
   pgd_iterations = attr.ib(default=1)  # 1 is the FGSM attack.
   pgd_epsilon = attr.ib(default=None)
+  random_init = attr.ib(default=False)
 
 
 @attr.s
@@ -95,7 +99,8 @@ def make_adv_reg_config(
     clip_value_min=attr.fields(AdvNeighborConfig).clip_value_min.default,
     clip_value_max=attr.fields(AdvNeighborConfig).clip_value_max.default,
     pgd_iterations=attr.fields(AdvNeighborConfig).pgd_iterations.default,
-    pgd_epsilon=attr.fields(AdvNeighborConfig).pgd_epsilon.default):
+    pgd_epsilon=attr.fields(AdvNeighborConfig).pgd_epsilon.default,
+    random_init=attr.fields(AdvNeighborConfig).random_init.default):
   """Creates an `nsl.configs.AdvRegConfig` object.
 
   Args:
@@ -124,6 +129,8 @@ def make_adv_reg_config(
       (FGSM) attack.
     pgd_epsilon: radius of the epsilon ball to project back to. Only used in
       Projected Gradient Descent (PGD) attack.
+    random_init: Apply a random perturbation before FGSM/PGD steps. Default set
+      to `False` for no random initialization being applied.
 
   Returns:
     An `nsl.configs.AdvRegConfig` object.
@@ -137,7 +144,8 @@ def make_adv_reg_config(
           clip_value_min=clip_value_min,
           clip_value_max=clip_value_max,
           pgd_iterations=pgd_iterations,
-          pgd_epsilon=pgd_epsilon))
+          pgd_epsilon=pgd_epsilon,
+          random_init=random_init))
 
 
 class AdvTargetType(enum.Enum):
@@ -291,6 +299,42 @@ class VirtualAdvConfig(object):
   approx_difference = attr.ib(default=1e-6)
 
 
+@attr.s
+class GraphBuilderConfig(object):
+  """Encapsulates configuration parameters for building a graph.
+
+  For more information, see `nsl.tools.build_graph_from_config`.
+
+  Attributes:
+    id_feature_name: The name of the feature in the input `tf.train.Example`
+      objects representing the ID of examples. Defaults to `'id'`.
+    embedding_feature_name: The name of the feature in the input
+      `tf.train.Example` objects representing the embedding of examples.
+      Defaults to `'embedding'`.
+    similarity_threshold: Threshold used to determine which edges to retain in
+      the resulting graph. Defaults to 0.8.
+    lsh_splits: On each LSH bucketing round, the space containing the input
+      instances will be randomly split/partitioned this many times for better
+      graph builder performance. See the `nsl.tools.build_graph_from_config`
+      documentation for details. Defaults to 0, in which case all pairs of
+      inputs will be compared, probably resulting in slow running times on
+      larger input sets.
+    lsh_rounds: The number of rounds of LSH bucketing to perform when
+      `lsh_splits > 0`. This is also the number of LSH buckets each point will
+      be hashed into. Defaults to 1.
+    random_seed: Value used to seed the random number generator used to perform
+      randomized LSH bucketing of the inputs when `lsh_splits > 0`. By default,
+      the generator will be initialized randomly, but setting this to any
+      integer will initialize it deterministically. Defaults to `None`.
+  """
+  id_feature_name = attr.ib(default='id')
+  embedding_feature_name = attr.ib(default='embedding')
+  similarity_threshold = attr.ib(default=0.8)
+  lsh_splits = attr.ib(default=0)
+  lsh_rounds = attr.ib(default=1)
+  random_seed = attr.ib(default=None)
+
+
 @attr.s
 class GraphNeighborConfig(object):
   """Specifies neighbor attributes for graph regularization.
