@@ -21,9 +21,10 @@ import collections
 import functools
 import types
 
+
 import attr
 import neural_structured_learning.configs as nsl_configs
-import neural_structured_learning.lib as nsl_lib
+from neural_structured_learning.lib import adversarial_neighbor
 import six
 import tensorflow as tf
 
@@ -87,7 +88,7 @@ def adversarial_loss(features,
     optimizer.apply_gradients(zip(gradients, model.trainable_variables))
   ```
 
-  Arguments:
+  Args:
     features: Input features, should be a `Tensor` or a collection of `Tensor`
       objects. If it is a collection, the first dimension of all `Tensor`
       objects inside should be the same (i.e. batch size).
@@ -95,7 +96,7 @@ def adversarial_loss(features,
     model: A callable that takes `features` as inputs and computes `predictions`
       as outputs. An example would be a `tf.keras.Model` object.
     loss_fn: A callable which calcualtes labeled loss from `labels`,
-      `predictions`, and `sample_weights`. An example would be a
+      `predictions`, and `sample_weight`. An example would be a
       `tf.keras.losses.Loss` object.
     sample_weights: (optional) A 1-D `Tensor` of weights for the examples, with
       the same length as the first dimension of `features`.
@@ -134,14 +135,15 @@ def adversarial_loss(features,
       predictions = predictions if predictions is not None else model(features)
       labeled_loss = loss_fn(labels, predictions, sample_weights)
 
-  adv_input, adv_sample_weights = nsl_lib.gen_adv_neighbor(
+  adv_input, adv_sample_weights = adversarial_neighbor.gen_adv_neighbor(
       features,
       labeled_loss,
       config=adv_config.adv_neighbor_config,
       gradient_tape=gradient_tape,
       pgd_model_fn=model,
-      pgd_loss_fn=functools.partial(loss_fn, sample_weights=sample_weights),
-      pgd_labels=labels)
+      pgd_loss_fn=functools.partial(loss_fn, sample_weight=sample_weights),
+      pgd_labels=labels,
+      use_while_loop=True)
   adv_output = model(adv_input)
   if sample_weights is not None:
     adv_sample_weights = tf.math.multiply(sample_weights, adv_sample_weights)
@@ -233,7 +235,7 @@ class _LossWrapper(tf.keras.losses.Loss):
 def _prepare_loss_fns(loss, output_names):
   """Converts `loss` to a list of per-output loss functions or objects."""
   # losses for multiple outputs indexed by name
-  if isinstance(loss, collections.Mapping):
+  if isinstance(loss, collections.abc.Mapping):
     for name in output_names:
       if name not in loss:
         raise ValueError(
@@ -245,7 +247,7 @@ def _prepare_loss_fns(loss, output_names):
     return [tf.keras.losses.get(loss) for _ in output_names]
 
   # losses for multiple outputs indexed by position
-  if isinstance(loss, collections.Sequence):
+  if isinstance(loss, collections.abc.Sequence):
     if len(loss) != len(output_names):
       raise ValueError('`loss` should have the same number of elements as '
                        'model output')
@@ -260,13 +262,13 @@ def _prepare_loss_weights(loss_weights, output_names):
   if loss_weights is None:
     return [1.0] * len(output_names)
 
-  if isinstance(loss_weights, collections.Sequence):
+  if isinstance(loss_weights, collections.abc.Sequence):
     if len(loss_weights) != len(output_names):
       raise ValueError('`loss_weights` should have the same number of elements '
                        'as model output')
     return list(map(float, loss_weights))
 
-  if isinstance(loss_weights, collections.Mapping):
+  if isinstance(loss_weights, collections.abc.Mapping):
     for name in output_names:
       if name not in loss_weights:
         raise ValueError('Loss weight for {} not found in `loss_weights` '
@@ -324,13 +326,13 @@ def _prepare_metric_fns(metrics, output_names, loss_wrappers):
   if metrics is None:
     return [[] for _ in output_names]
 
-  if not isinstance(metrics, (list, collections.Mapping)):
+  if not isinstance(metrics, (list, collections.abc.Mapping)):
     raise TypeError('`metrics` must be a list or a dict, got {}'.format(
         str(metrics)))
 
   to_list = lambda x: x if isinstance(x, list) else [x]
 
-  if isinstance(metrics, collections.Mapping):
+  if isinstance(metrics, collections.abc.Mapping):
     # Converts `metrics` from a dictionary to a list of lists using the order
     # specified in `output_names`.
     metrics = [to_list(metrics.get(name, [])) for name in output_names]
@@ -591,11 +593,11 @@ class AdversarialRegularization(tf.keras.Model):
       output_names = ['output_%d' % i for i in range(1, num_output + 1)]
     return output_names
 
-  def _compute_total_loss(self, labels, outputs, sample_weights=None):
+  def _compute_total_loss(self, labels, outputs, sample_weight=None):
     # `None` is passed instead of the actual metrics in order to skip computing
     # metric values and updating metric states.
     loss, _ = _compute_loss_and_metrics(self._labeled_losses, None, labels,
-                                        outputs, sample_weights)
+                                        outputs, sample_weight)
     return loss
 
   def _extract_labels_and_weights(self, inputs):
@@ -617,16 +619,21 @@ class AdversarialRegularization(tf.keras.Model):
     }
 
   def _call_base_model(self, inputs, **kwargs):
-    base_input_names = getattr(self.base_model, 'input_names', [])
+    base_input_names = getattr(self.base_model, 'input_names', None)
     if (isinstance(self.base_model, tf.keras.Sequential) and
+        base_input_names is not None and
         not set(base_input_names) & set(inputs.keys())):
       # In some cases, Sequential models are automatically compiled to graph
       # networks with automatically generated input names. In this case, the
       # user isn't expected to know those names, so we just flatten the inputs.
-      # But the input names are sometimes meaningful (e.g. DenseFeatures layer).
-      # We check if there is any intersection between the user-provided names
-      # and model's input names. If there is, we assume the names are meaningful
-      # and do name-based lookup in the next branch.
+      # But the input names are sometimes meaningful (e.g. explicitly set in
+      # `tf.keras.Input`). We check if there is any intersection between the
+      # user-provided names and model's input names. If there is, we assume the
+      # names are meaningful and do name-based lookup in the next branch. There
+      # are also cases that Sequential models aren't automatically compiled
+      # (e.g. with a DenseFeatures layer). In this case, we don't do any
+      # flattening or lookup, but rely on the _base_with_labels_in_features flag
+      # to indicate the correct set of features for the base model.
       inputs = tf.nest.flatten(self._remove_labels_and_weights(inputs))
     elif self.base_model._is_graph_network:  # pylint: disable=protected-access
       if base_input_names:
@@ -642,7 +649,8 @@ class AdversarialRegularization(tf.keras.Model):
   def _forward_pass(self, inputs, labels, sample_weights, base_model_kwargs):
     """Runs the usual forward pass to compute outputs, loss, and metrics."""
     with tf.GradientTape() as tape:
-      tape.watch(tf.nest.flatten(inputs))
+      tape.watch([t for t in tf.nest.flatten(inputs)
+                  if adversarial_neighbor.is_differentiable_tensor(t)])
       outputs = self._call_base_model(inputs, **base_model_kwargs)
       # If the base_model is a subclassed model, its output_names are not
       # available before its first call. If it is a dynamic subclassed model,
@@ -681,14 +689,17 @@ class AdversarialRegularization(tf.keras.Model):
         labeled_loss=labeled_loss,
         gradient_tape=tape,
         model_kwargs=kwargs)
-    self.add_loss(self.adv_config.multiplier * adv_loss)
-    self.add_metric(adv_loss, name='adversarial_loss', aggregation='mean')
+    scaled_adv_loss = self.adv_config.multiplier * adv_loss
+    self.add_loss(scaled_adv_loss)
+    self.add_metric(
+        scaled_adv_loss, name='scaled_adversarial_loss', aggregation='mean')
     return outputs
 
   def save(self, *args, **kwargs):
-    raise NotImplementedError(
-        'Saving `AdversarialRegularization` models is currently not supported. '
-        'Consider using `save_weights` or saving the `base_model`.')
+    """Saves the base model. See base class for details of the interface."""
+    # Adversarial regularization doesn't introduce new model variables, so
+    # saving the base model can capture all variables in the model.
+    self.base_model.save(*args, **kwargs)
 
   def perturb_on_batch(self, x, **config_kwargs):
     """Perturbs the given input to generates adversarial examples.
@@ -716,14 +727,15 @@ class AdversarialRegularization(tf.keras.Model):
 
     config_kwargs = {k: v for k, v in config_kwargs.items() if v is not None}
     config = attr.evolve(self.adv_config.adv_neighbor_config, **config_kwargs)
-    adv_inputs, _ = nsl_lib.gen_adv_neighbor(
+    adv_inputs, _ = adversarial_neighbor.gen_adv_neighbor(
         inputs,
         labeled_loss,
         config=config,
         gradient_tape=tape,
         pgd_model_fn=self._call_base_model,
         pgd_loss_fn=self._compute_total_loss,
-        pgd_labels=labels)
+        pgd_labels=labels,
+        use_while_loop=True)
 
     if tf.executing_eagerly():
       # Converts `Tensor` objects to NumPy arrays and keeps other objects (e.g.
