@@ -90,6 +90,23 @@ class UtilsTest(tf.test.TestCase, parameterized.TestCase):
     expected_tensor = tf.constant(0.0, shape=[2, 3])
     self.assertAllEqual(projected_tensor_dict['f1'], expected_tensor)
 
+  def testProjectToBallL2SingleTensor(self):
+    tensor = tf.constant([[3.0, -4.0], [-0.7, 2.4]])
+    projected_tensor = self.evaluate(
+        utils.project_to_ball(tensor, 1.0, configs.NormType.L2))
+    # norm: [5.0, 2.5], scale: [0.2, 0.4]
+    expected_result = [[0.6, -0.8], [-0.28, 0.96]]
+    self.assertAllClose(projected_tensor, expected_result)
+
+  def testProjectToBallL2TensorList(self):
+    tensors = [tf.constant([[1.0, -2.0], [-8.0, 1.0]]),
+               tf.constant([[2.0], [-4.0]])]
+    projected_tensors = self.evaluate(
+        utils.project_to_ball(tensors, 0.9, configs.NormType.L2))
+    # norm: [3.0, 0.9], scale: [0.3, 0.1]
+    expected_results = [[[0.3, -0.6], [-0.8, 0.1]], [[0.6], [-0.4]]]
+    self.assertAllClose(projected_tensors, expected_results)
+
   # The 3 test cases for this are as follows: (1) normalize both components. (2)
   # Clip components that exceed the radius, but preserve others. (3) Both
   # samples are within the ball, do nothing.
@@ -114,6 +131,57 @@ class UtilsTest(tf.test.TestCase, parameterized.TestCase):
     self.assertAllEqual(projected_tensor_dict['f1'], expected_f1)
     self.assertAllEqual(projected_tensor_dict['f2'], expected_f2)
 
+  def testProjectToBallLInfSingleTensor(self):
+    tensor = tf.constant([[1.0, -3.0], [-2.0, 4.0]])
+    projected_tensor = self.evaluate(
+        utils.project_to_ball(tensor, 2.5, configs.NormType.INFINITY))
+    expected_result = [[1.0, -2.5], [-2.0, 2.5]]
+    self.assertAllClose(projected_tensor, expected_result)
+
+  def testProjectToBallLInfTensorList(self):
+    tensors = [tf.constant([[1.0, -3.0], [-2.0, 4.0]]),
+               tf.constant([[0.0], [-5.0]])]
+    projected_tensors = self.evaluate(
+        utils.project_to_ball(tensors, 2.5, configs.NormType.INFINITY))
+    expected_results = [[[1.0, -2.5], [-2.0, 2.5]], [[0.0], [-2.5]]]
+    self.assertAllClose(projected_tensors, expected_results)
+
+  def testRandomInNormBallLInf(self):
+    tensors = {
+        'feature1': tf.constant([[.1], [.2]]),
+        'feature2': tf.constant([[[.3, .4], [.5, .6]], [[.7, .8], [.9, 1.]]]),
+    }
+    radius = 0.5
+    samples = self.evaluate(
+        utils.random_in_norm_ball(tensors, radius, configs.NormType.INFINITY))
+    self.assertSameElements(tensors.keys(), samples.keys())
+    flat_samples = tf.nest.flatten(samples)
+    for tensor, sample in zip(tf.nest.flatten(tensors), flat_samples):
+      self.assertShapeEqual(sample, tensor)
+      self.assertAllInRange(sample, -radius, radius)
+
+  @parameterized.named_parameters(
+      ('L2', configs.NormType.L2, 2),
+      ('L1', configs.NormType.L1, 1),
+  )
+  def testRandomInNormBall(self, norm_type, order):
+    tensors = {
+        'feature1': tf.constant([[.1], [.2]]),
+        'feature2': tf.constant([[[.3, .4], [.5, .6]], [[.7, .8], [.9, 1.]]]),
+    }
+    radius = 0.5
+    samples = self.evaluate(
+        utils.random_in_norm_ball(tensors, radius, norm_type))
+    self.assertSameElements(tensors.keys(), samples.keys())
+    flat_samples = tf.nest.flatten(samples)
+    for tensor, sample in zip(tf.nest.flatten(tensors), flat_samples):
+      self.assertShapeEqual(sample, tensor)
+    per_feature_norm = [
+        np.linalg.norm(sample, order, tuple(range(1, len(sample.shape))))
+        for sample in flat_samples]
+    global_norm = np.linalg.norm(np.stack(per_feature_norm, axis=1), order, 1)
+    self.assertAllLessEqual(global_norm, radius)
+
   def testNormalizeInfWithOnes(self):
     target_tensor = tf.constant(1.0, shape=[2, 4])
     normalized_tensor = self.evaluate(
